indexing

	description: "Class that covers a portable directory. Main use is to be %
	%able to loop through a single directory."

	author: "Berend de Boer, Marcio Marchini"
	date: "$Date: 2003/01/09 $";
	revision: "$Revision: #4 $"


deferred class

	ABSTRACT_DIRECTORY


inherit

	STDC_BASE

	EPX_FACTORY
		export
			{NONE} all
		end


feature -- Initialization

	make (a_directory_name: STRING) is
		require
			directory_name_not_empty: a_directory_name /= Void and then not a_directory_name.is_empty
		do
			create directory_name.make_from_string (a_directory_name)
			if my_item = Void then
				my_item := sh.make_with_capacity (max_filename_length)
			end
		end


feature -- Access

	close is
			-- Close directory entry (save resources now, don't wait for
			-- garbage collection). If you call start it will automatically
			-- reopen.
			-- If forth ends with exhausted, close is automatically called.
		require
			not_closed: is_open
		do
			safe_call (abstract_closedir (dirp))
			dirp := default_pointer
			my_item.wipe_out
			my_status := Void
			my_is_dot := False
			my_is_dotdot := False
		ensure
			closed: not is_open
			item_cleared: my_item.is_empty
			status_cleared: my_status = Void
		end

	start is
			-- Start directory traversal.
		do
			exhausted := False
			subdirectory := Void
			my_item.wipe_out
			if dirp = default_pointer then
				dirp := abstract_opendir (sh.string_to_pointer (directory_name))
				if dirp = default_pointer then
					raise_posix_error
				end
			else
				abstract_rewinddir (dirp)
			end
			forth
			first := True
		end

	forth is
		require
			opened: is_open
			not_exhausted: not exhausted
		local
			stop: BOOLEAN
			optional_status: ABSTRACT_STATUS
		do
			first := False
			if filter = Void then
				forth_recursive
			else
				from
					stop := False
				until
					stop
				loop
					forth_recursive
					if not exhausted and then filter.require_status then
						optional_status := status
					end
					stop :=
						exhausted or else
						filter.is_valid (
							optional_status,
							full_name)
				end
			end
		ensure
			no_handle_leak: not is_open implies dirp = default_pointer
		end


feature -- Influence browsing

	filter: ABSTRACT_PATH_FILTER

	set_filter (a_filter: ABSTRACT_PATH_FILTER) is
			-- Change the filter which determines which entries are returned.
		do
			filter := a_filter
		ensure
			filter_set: filter = a_filter
		end

	set_extension_filter (extension: STRING) is
			-- Create and set filter, `extension' is like ".e".
			-- `extension' shouldn't contain any wildcards
		require
			extension_not_empty: extension /= Void and then not extension.is_empty
			directory_browsing_not_in_progress: not is_open
		local
			ext_filter: EPX_EXTENSION_FILTER
		do
			create ext_filter.make (extension)
			set_filter (ext_filter)
		ensure
			extension_filter_set: filter /= Void
		end

	recursive: BOOLEAN
			-- Enable depth first directory browsing.
			-- Recursion doesn't work well with hard or symbolic links. Use
			-- SUS_DIRECTORY instead.

	set_recursive (enable: BOOLEAN) is
		do
			recursive := enable
		ensure
			recursive_set: recursive = enable
		end


feature -- Public state

	directory_name: STDC_PATH

	first: BOOLEAN

	full_name: STDC_PATH is
			-- combination of directory_name and item.
		require
			opened: is_open
		do
			if subdirectory = Void then
				if my_full_name = Void then
					create my_full_name.make (256)
				else
					my_full_name.wipe_out
				end
				Result := my_full_name
				Result.append_string (directory_name)
				Result.append_character (fs.path_separator)
				Result.append_string (my_item)
			else
				Result := subdirectory.full_name
			end
		ensure
			not_void: Result /= Void
		end

	item: STRING is
			-- The current entry, i.e. the path name.
			-- Even for recursive browsing, this entry is just the path
			-- name without any preceding sub directory name.
		require
			opened: is_open
		do
			if subdirectory = Void then
				Result := my_item
			else
				Result := subdirectory.item
			end
		end

	is_dot: BOOLEAN is
			-- True if `item' is equal to "."
		do
			if subdirectory = Void then
				Result := my_is_dot
			else
				Result := subdirectory.is_dot
			end
		end

	is_dotdot: BOOLEAN is
			-- True if `item' is equal to ".."
		do
			if subdirectory = Void then
				Result := my_is_dotdot
			else
				Result := subdirectory.is_dotdot
			end
		end

	status: ABSTRACT_STATUS_PATH is
			-- Return the status object for `item'.
		do
			if subdirectory = Void then
				if my_status = Void then
					my_status := fs.status (full_name)
				end
				Result := my_status
			else
				Result := subdirectory.status
			end
		end


feature -- Status report

	exhausted: BOOLEAN
			-- no more entries in this directory

	is_open: BOOLEAN is
			-- True if directory is ready for traversal
		do
			Result := dirp /= default_pointer
		end


feature -- Other directory related queries

	max_filename_length: INTEGER is
			-- maximum length of a file in this directory
		deferred
		end


feature -- Annihilation

	dispose is
		local
			r: INTEGER
		do
			if dirp /= default_pointer then
				r := abstract_closedir (dirp)
			end
		end


feature {NONE} -- directory browsing primitives

	forth_recursive is
			-- Next entry, depth-first recursive if needed.
		require
			opened: is_open
			not_exhausted: not exhausted
		local
			fn: STDC_PATH
		do
			if subdirectory /= Void then
				subdirectory.forth
				if subdirectory.exhausted then
					subdirectory := Void
				end
			else
				forth_single
				if recursive and not exhausted then
					if not is_dot and not is_dotdot then
						fn := full_name
						if my_status = Void then
							my_status := fs.status (fn)
						else
							my_status.set_filename (fn)
						end
						if my_status.is_directory then
							subdirectory := fs.browse_directory (fn)
							subdirectory.set_recursive (True)
							subdirectory.start
							subdirectory.forth
						end
					end
				else
					my_status := Void
				end
			end
		ensure
			my_status_in_sync: my_status /= Void implies my_status.path.is_equal (full_name)
		end

	forth_single is
			-- Go to next entry. Set `is_dot' and `is_dotdot'.
		require
			opened: is_open
			not_exhausted: not exhausted
		local
			dirent: POINTER
		do
			errno.clear
			dirent := abstract_readdir (dirp)
			if dirent = default_pointer then
				if errno.is_not_ok then
					raise_posix_error
				else
					exhausted := True
					close
				end
			else
				sh.set_string_from_pointer (my_item, abstract_d_name (dirent))
				my_is_dot :=
					my_item.count = 1 and then my_item.item (1) = '.'
				my_is_dotdot :=
					not my_is_dot and then
					(my_item.count = 2 and then
					 my_item.item (1) = '.' and then
					 my_item.item (2) = '.')
			end
		end


feature {NONE} -- Private state, does not take recursive browsing into account.

	dirp: POINTER
			-- Internal directory handle.

	my_full_name: STDC_PATH
			--- Cache for full name, avoid string allocations.
			-- This cache speeds up simple directory browsing programs
			-- which output `full_name' enormously to a factor 3.

	my_item: STRING
			-- The current entry, i.e. the path name.

	my_is_dot: BOOLEAN
			-- If `my_item' is equal to ".".

	my_is_dotdot: BOOLEAN
			-- If `my_item' is equal to "..".

	my_status: ABSTRACT_STATUS_PATH
			-- Optional status of current entry.

	subdirectory: ABSTRACT_DIRECTORY
			-- Used for recursive directory browsing.


feature {NONE} -- abstract API

	abstract_closedir (a_dirp: POINTER): INTEGER is
			-- Ends directory read operation
		require
			valid_dirp: a_dirp /= default_pointer
		deferred
		end

	abstract_opendir (a_dirname: POINTER): POINTER is
			-- Opens a directory
		require
			valid_dirp: a_dirname /= default_pointer
		deferred
		end

	abstract_readdir (a_dirp: POINTER): POINTER is
			-- Reads a directory
		require
			valid_dirp: a_dirp /= default_pointer
		deferred
		end

	abstract_rewinddir (a_dirp: POINTER) is
			-- Resets the readdir pointer
		require
			valid_dirp: a_dirp /= default_pointer
		deferred
		end

	abstract_d_name (a_dirent: POINTER): POINTER is
		require
			valid_dirp: a_dirent /= default_pointer
		deferred
		end


invariant

	dirp_remains_valid: is_open implies dirp /= default_pointer
	directory_name_not_empty: directory_name /= Void and then not directory_name.is_empty
	my_item_not_void: my_item /= Void
	valid_is_dot: my_is_dot = my_item.is_equal (".")
	valid_is_dotdot: my_is_dotdot = my_item.is_equal ("..")
	my_status_tracks_item: my_status /= Void implies my_status.path.is_equal (full_name)

end
