indexing

	description: "String utilities such as converting strings%
	%to pointers and vice versa."

	thanks: "The mico/E team for the idea."

	date: "$Date: 2003/01/09 $"
	revision: "$Revision: #3 $"

deferred class

	ABSTRACT_STRING_HELPER

inherit

	KL_IMPORTED_STRING_ROUTINES


feature -- Contracts for platform specific STRING_HELPER

	string_to_pointer (s: STRING): POINTER is
		require
			-- any `s', including Void
		deferred
		ensure
			void_result: s = Void implies Result = default_pointer
		end

	pointer_to_string (p: POINTER): STRING is
			-- returns a new string
		require
			-- any `p' including default_pointer
		deferred
		ensure
			empty_string: p = default_pointer implies equal(Result, "")
		end

	set_string_from_pointer (s: STRING; p: POINTER) is
			-- copies contents of p to s
		require
			valid_s: s /= Void
		deferred
		ensure
			empty_string: p = default_pointer implies s.count = 0
		end

feature {NONE} -- String to pointer cache

	converted_string_cache: ARRAY [STRING] is
			-- For some systems it is documented that a
			-- `string_to_pointer' conversion is unsafe when garbage
			-- collection is enabled. This utility stores up to 10
			-- converted strings, so the GC finds that they are still
			-- references, so it won't remove the string while a pointer
			-- is still out there.
		once
			create Result.make (0, 9)
		end

	index: INTEGER
			-- index in `converted_string_cache', to be used by `store_string'.

	store_string (s: STRING) is
			-- Cache string in `converted_string_cache'
		require
			s_not_void: s /= Void
		do
			converted_string_cache.put (s, index)
			if index = converted_string_cache.upper then
				index := converted_string_cache.lower
			else
				index := index + 1
			end
		end

feature -- String utilities that have different names per compiler...

	frozen clear (s: STRING) is
		obsolete
			"Please use s.wipe_out"
		do
			s.wipe_out
		ensure
			really_empty: s.count = 0
		end

	insert_character (s: STRING; c: CHARACTER; i: INTEGER) is
		obsolete "Please inherit from KL_IMPORTED_STRING_ROUTINES and use .STRING_.insert_character instead."
		require
			s_not_void: s /= Void
			valid_index: i >= 1 and i <= s.count + 1
		do
			STRING_.insert_character (s, c, i)
		end

	make_with_capacity (capacity: INTEGER): STRING is
			-- Create a new string.
		require
			valid_capacity: capacity >= 0
		deferred
		ensure
			have_result: Result /= Void
		end


feature -- general string utilities

	append_integer (s: STRING; i: INTEGER; width: INTEGER) is
			-- Append `i' to `s', pad until `width' with 0 if necessary.
		require
			s_not_void: s /= Void
			positive_integer: i >= 0
			valid_width: width >= 1
			i_fits_in_width: i.out.count <= width
		local
			t: STRING
			j: INTEGER
			count: INTEGER
		do
			t := i.out
			from
				j := width
				count := t.count
			until
				j = count
			loop
				s.append_character ('0')
				j := j - 1
			end
			s.append_string (t)
		ensure
			s_is_longer: s.count = old s.count + width
		end

	chop (s: STRING) is
			-- remove newline character (if any) from `s'
			-- only when the last characters from last_string are %N or
			-- %R%N, they're removed.
		require
			s_not_void: s /= Void
		do
			if s.count > 0 and then s.item (s.count) = '%N' then
				s.remove (s.count)
				if s.count > 0 and then s.item (s.count) = '%R' then
					s.remove (s.count)
				end
			end
		end

	has_substring (s, substring: STRING): BOOLEAN is
			-- True if `s' contains the string `substring'
		require
			s_not_void: s /= Void
			substring_not_empty: substring /= Void and then not substring.is_empty
		do
			Result := s.substring_index (substring, 1) /= 0
		end

	make_spaces (count: INTEGER): STRING is
			-- return string with `count' spaces
		require
			valid_count: count >= 0
		local
			i: INTEGER
		do
			Result := make_with_capacity (count)
			from
				i := 1
			until
				i > count
			loop
				Result.append_character (' ')
				i := i + 1
			end
		ensure
			enough_spaces: Result.count = count
		end

	split_on (s: STRING; on: CHARACTER): ARRAY [STRING] is
			-- Split `s' on a given character `on'.
			-- s = Void is treated as s.is_empty, in that case the Result
			-- doesn't contain any entries.
		local
			p, start: INTEGER
			tmp_string: STRING
			count: INTEGER
			large_array: ARRAY [STRING]
			i: INTEGER
		do
			if s /= Void and then not s.is_empty then
				-- attempt to avoid allocs
				create Result.make (0, 32)
				from
					start := 1
					p := s.index_of (on, start)
				variant
					(s.count + 1) - start
				until
					p = 0
				loop
					tmp_string := s.substring (start, p-1)
					Result.force (tmp_string, count)
					count := count + 1
					start := p + 1
					p := s.index_of (on, start)
				end

				if start <= s.count then
					tmp_string := s.substring (start, s.count)
					Result.force (tmp_string, count)
					count := count + 1
				elseif start = s.count + 1 then
					-- make sure that when ended with `on' we append an
					-- empty string
					Result.force ("", count)
					count := count + 1
				end
				Result.resize (0, count-1)
				if Result.upper /= count - 1 then
					large_array := Result
					create Result.make (0, count - 1)
					from
						i := 0
					until
						i > count - 1
					loop
						Result.put (large_array.item (i), i)
						i := i + 1
					end
				end
			else
				create Result.make (0, -1)
			end
		ensure
			array_empty_if_s_empty:
				(s = Void implies Result.count = 0) and
				(s.is_empty implies Result.count = 0)
			got_all_parts:
				s /= Void and then not s.is_empty implies
					Result.count = s.occurrences (on) + 1
			-- and for_each s element_of Result it_holds s /= Void
		end

	trim (s: STRING) is
			-- Remove leading and trailing white space. Actual definition
			-- of white space probably depends on vendors `left_adjust'
			-- and `right_adjust' implementations.
		require
			s_not_void: s /= Void
		do
			s.left_adjust
			s.right_adjust
		ensure
			s_same_size_or_smaller: s.count <= old s.count
		end


end
